import abc
import json
import math
import re
import sqlite3
from cpe import CPE

ECOSYSTEMS = {
    'rubygems': 'gem',
    'maven': 'java',
    'pip': 'python'
}
NAMESPACE_MAP_V4 = {
    'gem': 'ruby',
    'composer': 'php',
    'npm': 'javascript',
    'nuget': 'dotnet'
}
LINUX_FLAVORS = {
    'alpine': {
        'version_format': 'apk',
        'vendor': 'Alpine'
    },
    'amzn': {
        'version_format': 'rpm',
        'vendor': 'Amazon'
    },
    'centos': {
        'version_format': 'rpm',
        'vendor': 'CentOS'
    },
    'debian': {
        'version_format': 'dpkg',
        'vendor': 'Debian'
    },
    'ol': {
        'version_format': 'rpm',
        'vendor': 'Oracle'
    },
    'opensuse': {
        'version_format': 'rpm',
        'vendor': 'OPENSUSE'
    },
    'photon': {
        'version_format': 'rpm',
        'vendor': 'VMware'
    },
    'rhel': {
        'version_format': 'rpm',
        'vendor': 'Redhat'
    },
    'sles': {
        'version_format': 'rpm',
        'vendor': 'SLES'
    },
    'ubuntu': {
        'version_format': 'dpkg',
        'vendor': 'Ubuntu'
    }
}
LINUX_SEVERITY_MAP = {
    'low': 'Low',
    'medium': 'Medium',
    'high': 'High',
    'critical': 'Critical',
    'moderate': 'Medium',
    'important': 'High',
    'not yet assigned': 'Unknown',
    'unimportant': 'Low',
    'n/a': 'Unknown',
    'negligible': 'Negligible',
    'untriaged': 'Unknown',
    'end-of-life': 'Unknown',
}
version_dict = {
    'v2': {'vuln_data': ["id", "record_source", "package_name", "namespace", "version_constraint", "version_format",
                         "cpes", "proxy_vulnerabilities", "fixed_in_version"],
           'vuln_meta_data': ["id", "record_source", "severity", "links", "description", "cvss_v2", "cvss_v3"]},
    'v3': {'vuln_data': ["id", "package_name", "namespace", "version_constraint", "version_format", "cpes",
                         "related_vulnerabilities", "fixed_in_versions", "fix_state", "advisories"],
           'vuln_meta_data': ["id", "namespace", "data_source", "record_source", "severity", "urls", "description",
                              "cvss"]},
    'v4': {'vuln_data': ["id", "package_name", "namespace", "version_constraint", "version_format", "cpes",
                         "related_vulnerabilities", "fixed_in_versions", "fix_state", "advisories", "created_at"],
           'vuln_meta_data': ["id", "namespace", "data_source", "record_source", "severity", "urls", "description",
                              "cvss"]},
    'v5': {
        'vuln_data': ["id", "package_name", "namespace", "package_qualifiers", "version_constraint", "version_format",
                      "cpes", "related_vulnerabilities", "fixed_in_versions", "fix_state", "advisories", "created_at"],
        'vuln_meta_data': ["id", "namespace", "data_source", "record_source", "severity", "urls", "description",
                           "cvss"]}
}

LINUX_NAMESPACE = {
    'ol': 'oracle:distro:oraclelinux',
    'amzn': 'amazon:distro:amazonlinux',
    'photon': 'vmware:distro:photon'
}

SEVERITY_DICT = {'Unknown': 0, 'Negligible': 1, 'Low': 2, 'Medium': 3, 'High': 4, 'Critical': 5}


class GrypeCollector:
    """
    GrypeCollector class
    """
    def __init__(self, sqlite_db_objects, allow_versions, record_source=None):
        """
        Constructor of Grype Collector
        :param sqlite_db_objects: sqlite_db_objects
        :param allow_versions: allow_versions
        """
        self.sqlite_db_objects = sqlite_db_objects
        self.allow_versions = allow_versions
        self.record_source = record_source

    def populate_feeds(self, lst_vuln, lst_vuln_metadata, version):
        """
        Populate the feeds into sqlite tables
        :param lst_vuln: vulnerability table data
        :param lst_vuln_metadata: vulnerability_metadata table date
        :param version: version
        """
        sqlite_db = self.sqlite_db_objects[version]
        if sqlite_db:
            sqlite_db.insert_records(table='vulnerability', records=lst_vuln)
            sqlite_db.insert_records(table='vulnerability_metadata', records=lst_vuln_metadata)
        else:
            raise ValueError(f'Sqlite connection is not created for the selected version: {version}')

    @abc.abstractmethod
    def executor(self):
        """
        Abstract method acts as executor based on feed instance.
        """
        raise NotImplementedError("Implement the grype executor")


class SqliteDatabase:
    """
    Base SQLite database class
    """
    def __init__(self, db_file_path):
        """
        Constructor
        :param db_file_path: db_file_path
        """
        self.sqlite_connection = sqlite3.connect(db_file_path)
        self.sqlite_cursor = self.sqlite_connection.cursor()

    def __call__(self, built_datetime, version):
        """
        Magic method called by class instance to create tables.
        :param built_datetime: built_datetime
        :param version: version
        """
        self.create_tables()

    @abc.abstractmethod
    def create_tables(self):
        """
        Abstract method for creating tables
        """
        raise NotImplementedError("Implement create_tables abstract method in sqlite database class.")

    @abc.abstractmethod
    def insert_records(self, table, records):
        """
        Abstract method to insert records.
        """
        raise NotImplementedError("Implement insert_records abstract method in sqlite database class.")


class SqliteDatabaseV2(SqliteDatabase):
    """
    SQLite database V2 class
    """

    def create_tables(self):
        """
        Method to create sqlite tables.
        """
        sql_query = """
           CREATE TABLE if not exists id(
               build_timestamp text,
               schema_version INTEGER
               )"""
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
           CREATE TABLE if not exists vulnerability(
               pk INTEGER PRIMARY KEY,
               id text,
               record_source text,
               package_name text,
               namespace text,
               version_constraint text,
               version_format text,
               cpes text,
               proxy_vulnerabilities text,
               fixed_in_version text
               )"""
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
               CREATE TABLE if not exists vulnerability_metadata(
                   id text,
                   record_source text,
                   severity text,
                   links text,
                   description text,
                   cvss_v2 text,
                   cvss_v3 text,
                   PRIMARY KEY(id, record_source)
                   )"""

        self.sqlite_cursor.execute(sql_query)

        sql_query = """
            CREATE INDEX get_vulnerability_index ON vulnerability(package_name, namespace)
        """
        self.sqlite_cursor.execute(sql_query)

        self.sqlite_connection.commit()

    def insert_records(self, table, records):
        """
        Method to insert bulk records in the sqlite table.
        :param table: SQL query
        :param records: data
        """
        if records:
            sql_query = {
                'id': "INSERT INTO id VALUES (?,?)",
                'vulnerability': "INSERT INTO vulnerability VALUES (NULL,?,?,?,?,?,?,?,?,?)",
                'vulnerability_metadata': "INSERT INTO vulnerability_metadata VALUES (?,?,?,?,?,?,?)"
            }
            self.sqlite_cursor.executemany(sql_query[table], records)
            self.sqlite_connection.commit()


class SqliteDatabaseV3(SqliteDatabase):
    """
    SQLite database V3 class
    """

    def create_tables(self):
        """
        Method to create sqlite tables.
        """
        sql_query = """
           CREATE TABLE if not exists id(
               build_timestamp text,
               schema_version INTEGER
               )"""
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
           CREATE TABLE if not exists vulnerability(
               pk INTEGER PRIMARY KEY,
               id text,
               package_name text,
               namespace text,
               version_constraint text,
               version_format text,
               cpes text,
               related_vulnerabilities text,
               fixed_in_versions text,
               fix_state text,
               advisories text
               )"""
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
               CREATE TABLE if not exists vulnerability_metadata(
                   id text,
                   namespace text,
                   data_source text,
                   record_source text,
                   severity text,
                   urls text,
                   description text,
                   cvss text,
                   PRIMARY KEY(id, namespace)
                   )"""

        self.sqlite_cursor.execute(sql_query)

        self.sqlite_cursor.execute(sql_query)

        sql_query = """
            CREATE INDEX get_vulnerability_index ON vulnerability (package_name, namespace)
        """
        self.sqlite_cursor.execute(sql_query)
        self.sqlite_connection.commit()

    def insert_records(self, table, records):
        """
        Method to insert bulk records in the sqlite table.
        :param table: SQL query
        :param records: data
        """
        if records:
            sql_query = {
                'id': "INSERT INTO id VALUES (?,?)",
                'vulnerability': "INSERT INTO vulnerability VALUES (NULL,?,?,?,?,?,?,?,?,?,?)",
                'vulnerability_metadata': "INSERT INTO vulnerability_metadata VALUES (?,?,?,?,?,?,?,?)"
            }
            self.sqlite_cursor.executemany(sql_query[table], records)
            self.sqlite_connection.commit()


class SqliteDatabaseV4(SqliteDatabase):
    """
    SQLite database V4 class
    """

    def create_tables(self):
        """
        Method to create sqlite tables.
        """
        sql_query = """
           CREATE TABLE if not exists id(
               build_timestamp text,
               schema_version INTEGER
               )"""
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
           CREATE TABLE if not exists vulnerability(
               pk INTEGER PRIMARY KEY,
               id text,
               package_name text,
               namespace text,
               version_constraint text,
               version_format text,
               cpes text,
               related_vulnerabilities text,
               fixed_in_versions text,
               fix_state text,
               advisories text,
               created_at text
               )"""
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
               CREATE TABLE if not exists vulnerability_metadata(
                   id text,
                   namespace text,
                   data_source text,
                   record_source text,
                   severity text,
                   urls text,
                   description text,
                   cvss text,
                   PRIMARY KEY(id, namespace)
                   )"""

        self.sqlite_cursor.execute(sql_query)

        sql_query = """
               CREATE TABLE if not exists vulnerability_match_exclusion(
                   pk INTEGER PRIMARY KEY,
                   id text,
                   constraints text DEFAULT null,
                   justification text
                   )"""

        self.sqlite_cursor.execute(sql_query)

        sql_query = """
            CREATE INDEX get_vulnerability_index ON vulnerability (package_name, namespace, created_at)
        """
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
            CREATE INDEX get_vulnerability_match_exclusion_index ON vulnerability_match_exclusion (id)
        """
        self.sqlite_cursor.execute(sql_query)
        self.sqlite_connection.commit()

    def insert_records(self, table, records):
        """
        Method to insert bulk records in the sqlite table.
        :param table: SQL query
        :param records: data
        """
        if records:
            sql_query = {
                'id': "INSERT INTO id VALUES (?,?)",
                'vulnerability': "INSERT INTO vulnerability VALUES (NULL,?,?,?,?,?,?,?,?,?,?,?)",
                'vulnerability_metadata': "INSERT INTO vulnerability_metadata VALUES (?,?,?,?,?,?,?,?)"
            }
            self.sqlite_cursor.executemany(sql_query[table], records)
            self.sqlite_connection.commit()


class SqliteDatabaseV5(SqliteDatabase):
    """
    SQLite database V5 class
    """

    def create_tables(self):
        """
        Method to create sqlite tables.
        """
        sql_query = """
           CREATE TABLE if not exists id(
               build_timestamp text,
               schema_version INTEGER
               )"""
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
           CREATE TABLE if not exists vulnerability(
               pk INTEGER PRIMARY KEY,
               id text,
               package_name text,
               namespace text,
               package_qualifiers text,
               version_constraint text,
               version_format text,
               cpes text,
               related_vulnerabilities text,
               fixed_in_versions text,
               fix_state text,
               advisories text,
               created_at text
               )"""
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
               CREATE TABLE if not exists vulnerability_metadata(
                   id text,
                   namespace text,
                   data_source text,
                   record_source text,
                   severity text,
                   urls text,
                   description text,
                   cvss text,
                   PRIMARY KEY(id, namespace)
                   )"""

        self.sqlite_cursor.execute(sql_query)

        sql_query = """
               CREATE TABLE if not exists vulnerability_match_exclusion(
                   pk INTEGER PRIMARY KEY,
                   id text,
                   constraints text DEFAULT null,
                   justification text
                   )"""

        self.sqlite_cursor.execute(sql_query)

        sql_query = """
            CREATE INDEX get_vulnerability_index ON vulnerability (package_name, namespace, created_at)
        """
        self.sqlite_cursor.execute(sql_query)

        sql_query = """
            CREATE INDEX get_vulnerability_match_exclusion_index ON vulnerability_match_exclusion (id)
        """
        self.sqlite_cursor.execute(sql_query)
        self.sqlite_connection.commit()

    def insert_records(self, table, records):
        """
        Method to insert bulk records in the sqlite table.
        :param table: SQL query
        :param records: data
        """
        if records:
            sql_query = {
                'id': "INSERT INTO id VALUES (?,?)",
                'vulnerability': "INSERT INTO vulnerability VALUES (NULL,?,?,?,?,?,?,?,?,?,?,?,?)",
                'vulnerability_metadata': "INSERT INTO vulnerability_metadata VALUES (?,?,?,?,?,?,?,?)"
            }
            self.sqlite_cursor.executemany(sql_query[table], records)
            self.sqlite_connection.commit()


def get_sqlite_db_object(version, db_file_path):
    """
    Function to get the appropriate sqlite instance to be use later.
    :param version: version
    :param db_file_path: db_file_path
    :return: sqlite database instance
    """
    sqlite_db = None
    if version == 'v2':
        sqlite_db = SqliteDatabaseV2(db_file_path)
    if version == 'v3':
        sqlite_db = SqliteDatabaseV3(db_file_path)
    if version == 'v4':
        sqlite_db = SqliteDatabaseV4(db_file_path)
    if version == 'v5':
        sqlite_db = SqliteDatabaseV5(db_file_path)
    return sqlite_db


def handle_nan(val):
    """
    This function replaces NaN values with None otherwise it returns same value.
    :param val: val
    :return: val
    """
    if (isinstance(val, float) and math.isnan(val)) or (val is None) or (isinstance(val, str) and val.lower() == 'nan'):
        return 0.0
    return val


def get_cpe_details(cpe_uri, is_application=False):
    """
    This function will get the CPE product details for the cpe_uri

    :param cpe_uri: cpe_uri for which the product details have to be fetched.
    :param is_application: check for cpe application products
    :return: cpe product details for cpe_uri
    """
    cpe_details = {}

    cpe_obj = CPE(cpe_uri)
    for part, value in cpe_obj.items():
        if not value:
            continue

        if part != 'app' and is_application:
            continue

        val = value[0]
        cpe_details['cpe_uri'] = cpe_uri
        cpe_details['part'] = part
        cpe_details['vendor'] = str(val['vendor'])
        cpe_details['product'] = str(val['product'])
        cpe_details['version'] = str(val['version']).replace('\\', '')
        cpe_details['update'] = str(val['update'])
        cpe_details['edition'] = str(val['edition'])
        cpe_details['language'] = str(val['language'])
        cpe_details['sw_edition'] = str(val['sw_edition'])
        cpe_details['target_sw'] = str(val['target_sw'])
        cpe_details['target_hw'] = str(val['target_hw'])
        cpe_details['other'] = str(val['other'])
        cpe_details['cpe_uri_2_3'] = str(cpe_obj.as_uri_2_3())

    return cpe_details


def get_version_constraints(cpe_product, version):
    """
    Function to get the version constraints of the cpe product
    :param cpe_product: cpe product name
    :param version: specified version number
    :return: returns the list of version contraints.
    """
    version_constraints = []
    version_start_including = cpe_product[1]
    version_end_including = cpe_product[3]
    version_start_excluding = cpe_product[2]
    version_end_excluding = cpe_product[4]
    if version_start_including != 'None':
        version_constraints.append('>= ' + version_start_including)
    if version_start_excluding != 'None':
        version_constraints.append('> ' + version_start_excluding)
    if version_end_including != 'None':
        version_constraints.append('<= ' + version_end_including)
    if version_end_excluding != 'None':
        version_constraints.append('< ' + version_end_excluding)

    if version_constraints:
        version_constraints = [', '.join(version_constraints)]
    if version and version != '<ANY>':
        version_constraints.append('= ' + version)
    return version_constraints if version_constraints else ['']


def initialise_variables():
    """
    Function to initialise variables
    :return: col_dict
    """
    col_dict = {'id': None,
                'record_source': None,
                'severity': None,
                'links': None, 'urls': None,
                'description': None,
                'cvss_v2': None, 'cvss_v3': None, 'cvss': '[]',
                'data_source': None,
                'package_name': None,
                'namespace': None,
                'version_constraint': None,
                'version_format': None,
                'cpes': "null", 'advisories': "null",
                'fix_state': None, 'created_at': None,
                'package_qualifiers': None, 'related_vulnerabilities': None,
                'fixed_in_version': None, 'fixed_in_versions': None}
    return col_dict


def get_linux_cvss_data(cve_details_recs, cve_id, allow_versions, severity):
    """
    Function to get linux cvss data
    :param cve_details_recs: cve_details_recs
    :param cve_id: cve_id
    :param allow_versions: allow_versions
    :param severity: severity
    :return: cvss
    """
    cvss = []
    if allow_versions in ('v3',):
        if cve_id in cve_details_recs:
            if cve_details_recs[cve_id][4] != '':
                cvss.append({
                    "VendorMetadata": {
                        "BaseSeverity": severity
                    },
                    "Metrics":
                        {
                            "BaseScore": handle_nan(cve_details_recs[cve_id][5]),
                            "ExploitabilityScore": handle_nan(cve_details_recs[cve_id][6]),
                            "ImpactScore": handle_nan(cve_details_recs[cve_id][7])
                        },
                    "Vector": cve_details_recs[cve_id][4],
                    "Version": '2.0',
                })
            elif cve_details_recs[cve_id][8] != '':
                cvss.append({
                    "VendorMetadata": {
                        "BaseSeverity": severity
                    },
                    "Metrics":
                        {
                            "BaseScore": handle_nan(cve_details_recs[cve_id][9]),
                            "ExploitabilityScore": handle_nan(cve_details_recs[cve_id][10]),
                            "ImpactScore": handle_nan(cve_details_recs[cve_id][11])
                        },
                    "Vector": cve_details_recs[cve_id][8],
                    "Version": cve_details_recs[cve_id][8].split(":")[1].split("/")[0],
                })
            else:
                cvss = []

    if allow_versions in ('v4', 'v5'):
        if cve_id in cve_details_recs:
            if cve_details_recs[cve_id][4] != '':
                cvss.append({
                    "vendor_metadata": {
                        "base_severity": severity
                    },
                    "metrics":
                        {
                            "base_score": handle_nan(cve_details_recs[cve_id][5]),
                            "exploitability_score": handle_nan(cve_details_recs[cve_id][6]),
                            "impact_score": handle_nan(cve_details_recs[cve_id][7])
                        },
                    "vector": cve_details_recs[cve_id][4],
                    "version": '2.0',
                })
            elif cve_details_recs[cve_id][8] != '':
                cvss.append({
                    "vendor_metadata": {
                        "base_severity": severity
                    },
                    "metrics":
                        {
                            "base_score": handle_nan(cve_details_recs[cve_id][9]),
                            "exploitability_score": handle_nan(cve_details_recs[cve_id][10]),
                            "impact_score": handle_nan(cve_details_recs[cve_id][11])
                        },
                    "vector": cve_details_recs[cve_id][8],
                    "version":  cve_details_recs[cve_id][8].split(":")[1].split("/")[0],
                })
            else:
                cvss = []

    return cvss


def get_nvd_cvss_data(cve_info, allow_versions):
    """
    Function to get linux cvss data
    :param cve_info: cve_info
    :param allow_versions: allow_versions
    :return: tuple
    """
    cvss = []
    severity = 'UNKNOWN'.title()
    cvss_v2_impact, cvss_v3_impact = {}, {}
    if allow_versions in ('v2', 'v3'):
        if cve_info['cvss_vector'] and cve_info['cvss_vector'] != '':
            cvss_v2_impact = {
                "VendorMetadata": None,
                "Metrics":
                    {
                        "BaseScore": handle_nan(cve_info['cvss_score']),
                        "ExploitabilityScore": handle_nan(cve_info['cvss_exploit_subscore']),
                        "ImpactScore": handle_nan(cve_info['cvss_impact_subscore'])
                    },
                "Vector": cve_info['cvss_vector'],
                "Version":  '2.0',
            }
            cvss.append(cvss_v2_impact)
            severity = cve_info['cvss_severity'].title() if cve_info['cvss_severity'] else severity

        if cve_info['cvss_v3_vector'] and cve_info['cvss_v3_vector'] != '':
            cvss_v3_impact = {
                "VendorMetadata": None,
                "Metrics":
                    {
                        "BaseScore": handle_nan(cve_info['cvss_v3_score']),
                        "ExploitabilityScore": handle_nan(cve_info['cvss_v3_exploitability_score']),
                        "ImpactScore": handle_nan(cve_info['cvss_v3_impact_score'])
                    },
                "Vector": cve_info['cvss_v3_vector'],
                "Version": cve_info['cvss_v3_vector'].split(":")[1].split("/")[0] if cve_info['cvss_v3_vector'] else "",
            }
            cvss.append(cvss_v3_impact)
            severity = cve_info['cvss_v3_severity'].title() if cve_info['cvss_v3_severity'] else severity
    if allow_versions in ('v4', 'v5'):
        if cve_info['cvss_vector'] and cve_info['cvss_vector'] != '':
            cvss.append({
                "vendor_metadata": None,
                "metrics":
                    {
                        "base_score": handle_nan(cve_info['cvss_score']),
                        "exploitability_score": handle_nan(cve_info['cvss_exploit_subscore']),
                        "impact_score": handle_nan(cve_info['cvss_impact_subscore'])
                    },
                "vector": cve_info['cvss_vector'],
                "version": '2.0',
            })
            severity = cve_info['cvss_severity'].title() if cve_info['cvss_severity'] else severity
        if cve_info['cvss_v3_vector'] and cve_info['cvss_v3_vector'] != '':
            cvss.append({
                "vendor_metadata": None,
                "metrics":
                    {
                        "base_score": handle_nan(cve_info['cvss_v3_score']),
                        "exploitability_score": handle_nan(cve_info['cvss_v3_exploitability_score']),
                        "impact_score": handle_nan(cve_info['cvss_v3_impact_score'])
                    },
                "vector": cve_info['cvss_v3_vector'],
                "version": cve_info['cvss_v3_vector'].split(":")[1].split("/")[0] if cve_info['cvss_v3_vector'] else "",
            })
            severity = cve_info['cvss_v3_severity'].title() if cve_info['cvss_v3_severity'] else severity
    return cvss, severity, cvss_v2_impact, cvss_v3_impact


class GrypeFeed:
    """
    GrypeFeed
    """

    def __init__(self):
        """
        Constructor of Grype Feed
        """
        self.created_dict = {}
        self.lst_vuln_metadata, self.lst_vuln_data, self.vulnmeta_data, self.vuln_data = [], [], {}, {}

    def get_github_data(self, data, allow_versions, cve_details_recs):
        """
        Method to get github data
        :param data: data
        :param allow_versions: allow_versions
        :param cve_details_recs: cve_details_recs
        :return: tuple
        """
        for row in data:
            cve_id = row['advisory_id']
            if row['deleted'] or (cve_id in cve_details_recs and cve_details_recs[cve_id]
                                  and '** REJECT **' in cve_details_recs[cve_id]):
                continue
            table_col_dict = initialise_variables()
            new_created_at = row['created_at'].date()
            last_created_at_key = row['advisory_id'] + "::" + row['package']
            if last_created_at_key not in self.created_dict:
                self.created_dict[last_created_at_key] = new_created_at

            if new_created_at > self.created_dict[last_created_at_key]:
                self.created_dict[last_created_at_key] = new_created_at

            ghsa_id = row['advisory_url'].split("/")[-1]
            table_col_dict['id'] = ghsa_id if ghsa_id.startswith("GHSA") else None
            if table_col_dict['id']:
                eco_sys = str(row['ecosystem']).lower()
                ecosystem = ECOSYSTEMS.get(eco_sys, eco_sys)
                table_col_dict['record_source'] = 'github:' + ecosystem
                table_col_dict['version_format'] = 'python' if ecosystem == 'python' else 'unknown'
                severity = 'Medium' if row['severity'].lower() == 'moderate' else row['severity'].title()
                table_col_dict['severity'] = severity.strip().title() if severity else 'Unknown'
                table_col_dict['links'] = json.dumps([row['advisory_url']])
                table_col_dict['description'] = row['summary']
                table_col_dict['package_name'] = str(row['package']).lower() if ecosystem == 'nuget' else row['package']
                table_col_dict['namespace'] = 'github:' + ecosystem
                table_col_dict['data_source'] = row['advisory_url']
                table_col_dict['urls'] = table_col_dict['links']
                table_col_dict['version_constraint'] = row['affected_version'].replace(" <", ",<").replace(" ", "")
                table_col_dict['proxy_vulnerabilities'] = json.dumps([table_col_dict['id']]) if table_col_dict[
                    'id'].startswith('CVE') else '[]'
                table_col_dict['fixed_in_version'] = row['fixed_version']
                if 'v3' in allow_versions:
                    table_col_dict['related_vulnerabilities'] = json.dumps([{"ID": cve_id, "Namespace": "nvd"}]).replace(
                        " ", "") if cve_id.startswith("CVE") else '[]' if str(cve_id).startswith('CVE') else '[]'
                else:
                    table_col_dict['related_vulnerabilities'] = json.dumps([{"ID": cve_id, "Namespace": "nvd:cpe"}]).replace(
                        " ", "") if cve_id.startswith("CVE") else '[]' if str(cve_id).startswith('CVE') else '[]'
                table_col_dict['fixed_in_versions'] = json.dumps([row['fixed_version']]) if row[
                    'fixed_version'] else 'null'
                table_col_dict['fix_state'] = 'fixed' if row['fixed_version'] else 'not-fixed'
                table_col_dict['created_at'] = str(self.created_dict[last_created_at_key])

                if allow_versions in ('v3', 'v4', 'v5'):
                    table_col_dict['record_source'] = 'github:github:' + ecosystem
                if allow_versions in ('v4', 'v5'):
                    namespace = NAMESPACE_MAP_V4[ecosystem] if ecosystem in NAMESPACE_MAP_V4 else ecosystem
                    table_col_dict['namespace'] = 'github:language:' + namespace

                vuln_metadata_dict, vuln_data_dict = {}, {}
                if allow_versions in version_dict:
                    meta_data_fields = version_dict[allow_versions]['vuln_meta_data']
                    for fields in meta_data_fields:
                        vuln_metadata_dict[fields] = table_col_dict[fields]
                    vuln_meta_rec = table_col_dict['id'] + "::" + table_col_dict['record_source']
                    if vuln_meta_rec not in self.vulnmeta_data:
                        self.vulnmeta_data[vuln_meta_rec] = 1
                        self.lst_vuln_metadata.append(tuple(vuln_metadata_dict.values()))

                    vuln_data_fields = version_dict[allow_versions]['vuln_data']
                    for fields in vuln_data_fields:
                        vuln_data_dict[fields] = table_col_dict[fields]
                    vuln_rec = table_col_dict['id'] + "::" + table_col_dict['namespace'] + "::" + table_col_dict[
                        'package_name'] + "::" + table_col_dict['version_constraint']
                    if vuln_rec not in self.vuln_data:
                        self.vuln_data[vuln_rec] = 1
                        self.lst_vuln_data.append(tuple(vuln_data_dict.values()))

        return self.lst_vuln_data, self.lst_vuln_metadata

    def get_linux_data(self, data, allow_versions, cve_data_dict, os_platform):
        """
        Method to get linux data
        :param data: data
        :param allow_versions: allow_versions
        :param cve_data_dict: cve_data_dict
        :param os_platform: os_platform
        :return: tuple
        """
        for rows in data:
            table_col_dict = initialise_variables()
            cve_id, package, fixed_version, advisory_severity, advisory_url, advisory_url_alias, description, version, created_at = rows

            new_created_at = created_at.date()
            last_created_at_key = cve_id + "::" + package
            if last_created_at_key not in self.created_dict:
                self.created_dict[last_created_at_key] = new_created_at

            if new_created_at > self.created_dict[last_created_at_key]:
                self.created_dict[last_created_at_key] = new_created_at

            data_source = advisory_url
            if os_platform in ('photon', 'alpine'):
                data_source = "http://cve.mitre.org/cgi-bin/cvename.cgi?name={}".format(cve_id)
            if os_platform in ('rhel',):
                data_source = 'https://access.redhat.com/security/cve/{}'.format(cve_id)

            links, cve_list = [], []
            if os_platform in ('ol', 'centos'):
                links.append(advisory_url)
                cve_list = cve_data_dict['urls_dict'].get(advisory_url, [])
                for cveid in set(cve_list):
                    url = ''
                    if cveid:
                        if os_platform == 'ol':
                            url = f'https://linux.oracle.com/cve/{cveid}.html'
                        if os_platform == 'centos':
                            url = f'https://access.redhat.com/security/cve/{cveid}'
                        links.append(url)

            if os_platform in ('amzn', 'ol', 'centos'):
                table_col_dict['id'] = advisory_url.split("/")[-1].replace(".html", "").strip()
            else:
                table_col_dict['id'] = cve_id

            if os_platform == 'photon':
                platform = os_platform + ':' + version + '.0'
            else:
                platform = os_platform + ':' + version

            if cve_id in cve_data_dict['cve_info']:
                if (not description) and cve_data_dict['cve_info']:
                    description = cve_data_dict['cve_info'][cve_id][1] if cve_data_dict['cve_info'][cve_id][1] else ''

                    if '** REJECT **' in description:
                        continue

                if (not advisory_severity) and cve_data_dict['cve_info'][cve_id]:
                    if cve_data_dict['cve_info'][cve_id][3]:
                        advisory_severity = cve_data_dict['cve_info'][cve_id][3]
                    elif cve_data_dict['cve_info'][cve_id][2]:
                        advisory_severity = cve_data_dict['cve_info'][cve_id][2]
                    else:
                        advisory_severity = 'Unknown'
            sev = advisory_severity.strip().lower() if advisory_severity else 'Unknown'
            severity = LINUX_SEVERITY_MAP[sev] if sev in LINUX_SEVERITY_MAP else 'Unknown'

            cvss = []
            if os_platform in ('rhel',):
                cvss = get_linux_cvss_data(cve_data_dict['cve_info'], cve_id, allow_versions, severity)

            table_col_dict['record_source'] = 'vulnerabilities:' + platform
            table_col_dict['severity'] = severity
            table_col_dict['links'] = json.dumps([advisory_url])
            table_col_dict['data_source'] = data_source
            table_col_dict['description'] = description if description and os_platform != 'ol' else ''
            table_col_dict['cvss_v2'] = None
            table_col_dict['cvss_v3'] = None
            table_col_dict['package_name'] = package
            table_col_dict['namespace'] = platform
            table_col_dict['version_format'] = LINUX_FLAVORS[os_platform]['version_format']
            table_col_dict['cpes'] = "null"
            table_col_dict['proxy_vulnerabilities'] = []
            table_col_dict['fixed_in_version'] = fixed_version
            table_col_dict['related_vulnerabilities'] = [table_col_dict['id']] if table_col_dict['id'].startswith(
                "CVE") else []
            table_col_dict['fix_state'] = 'fixed' if fixed_version else 'not-fixed'
            table_col_dict['advisories'] = 'null'
            table_col_dict['created_at'] = str(self.created_dict[last_created_at_key])
            table_col_dict['urls'] = json.dumps([data_source]) if os_platform not in ('ol', 'centos') else json.dumps(
                links)
            table_col_dict['cvss'] = json.dumps(cvss) if cvss else '[]'

            if os_platform == 'sles':
                fixed_version = "0:" + fixed_version if fixed_version != '' else ''
            table_col_dict['fixed_in_versions'] = json.dumps([fixed_version]) if fixed_version else 'null'
            table_col_dict['version_constraint'] = f'< {fixed_version}' if fixed_version != '' else ''
            if os_platform in ('rhel',):
                table_col_dict['fix_state'] = 'not-fixed' if not fixed_version and not table_col_dict[
                    'version_constraint'] \
                    else table_col_dict['fix_state']
                if advisory_url:
                    adv_id = advisory_url.split("/")[-1].replace(".html", "")
                    table_col_dict['advisories'] = json.dumps([{"ID": adv_id, "Link": advisory_url}]).replace(" ", "")
                else:
                    table_col_dict['advisories'] = 'null'

            if os_platform == 'debian':
                table_col_dict['fix_state'] = 'not-fixed' if not fixed_version and not table_col_dict[
                    'version_constraint'] else table_col_dict['fix_state']
                dsa_id = advisory_url_alias.split("/")[-1].replace(".html", "") if advisory_url_alias else "null"

                if advisory_url_alias:
                    table_col_dict['advisories'] = \
                        json.dumps([{"ID": dsa_id, "Link": advisory_url_alias}]).replace(" ", "")

            if allow_versions in ('v4', 'v5'):
                namespace_str = LINUX_NAMESPACE[os_platform] \
                    if os_platform in LINUX_NAMESPACE \
                    else f'{(cve_data_dict["vendor"]).lower()}:distro:{cve_data_dict["vendor"].lower()}'
                table_col_dict['namespace'] = namespace_str + ':' + version
                if os_platform == 'photon':
                    table_col_dict['namespace'] = namespace_str + ':' + version + '.0'

            vuln_metadata_dict, vuln_data_dict = {}, {}
            if allow_versions in version_dict:
                meta_data_fields = version_dict[allow_versions]['vuln_meta_data']
                for fields in meta_data_fields:
                    vuln_metadata_dict[fields] = table_col_dict[fields]
                vuln_meta_rec = table_col_dict['id'] + "::" + table_col_dict['record_source']
                if vuln_meta_rec not in self.vulnmeta_data:
                    self.vulnmeta_data[vuln_meta_rec] = {}
                else:
                    if SEVERITY_DICT[self.vulnmeta_data[vuln_meta_rec]['severity']] > SEVERITY_DICT[severity]:
                        continue
                self.vulnmeta_data[vuln_meta_rec]['severity'] = severity
                self.vulnmeta_data[vuln_meta_rec]['data'] = tuple(vuln_metadata_dict.values())

                vuln_data_fields = version_dict[allow_versions]['vuln_data']
                for fields in vuln_data_fields:
                    vuln_data_dict[fields] = table_col_dict[fields]
                vuln_rec = table_col_dict['id'] + "::" + table_col_dict['record_source'] + "::" + table_col_dict[
                    'package_name'] + "::" + table_col_dict['version_constraint']
                if allow_versions == 'v2':
                    if os_platform in ('amzn', 'ol'):
                        if vuln_rec not in self.vuln_data:
                            self.vuln_data[vuln_rec] = vuln_data_dict
                        if cve_id not in self.vuln_data[vuln_rec]['proxy_vulnerabilities']:
                            self.vuln_data[vuln_rec]['proxy_vulnerabilities'].append(cve_id)
                    else:
                        self.vuln_data[vuln_rec] = vuln_data_dict
                else:
                    if os_platform in ('amzn', 'ol', 'centos'):
                        if vuln_rec not in self.vuln_data:
                            self.vuln_data[vuln_rec] = vuln_data_dict
                        if cve_id not in self.vuln_data[vuln_rec]['related_vulnerabilities']:
                            self.vuln_data[vuln_rec]['related_vulnerabilities'].append(cve_id)
                    else:
                        self.vuln_data[vuln_rec] = vuln_data_dict

        # TODO: Refactor this code.
        lst_vuln = []
        if allow_versions == 'v2':
            for _, val in self.vuln_data.items():
                val['proxy_vulnerabilities'] = json.dumps(val['proxy_vulnerabilities'])
                lst_vuln.append(tuple(val.values()))
        if allow_versions == 'v3':
            for _, val in self.vuln_data.items():
                temp_list = []
                for cve in val['related_vulnerabilities']:
                    if cve:
                        temp_list.append({"ID": cve, "Namespace": "nvd"})
                val['related_vulnerabilities'] = json.dumps(temp_list).replace(" ", "")
                lst_vuln.append(tuple(val.values()))
        if allow_versions in ('v4', 'v5'):
            for _, val in self.vuln_data.items():
                temp_list = []
                for cve in val['related_vulnerabilities']:
                    if cve:
                        temp_list.append({"ID": cve, "Namespace": "nvd:cpe"})
                val['related_vulnerabilities'] = json.dumps(temp_list).replace(" ", "")
                lst_vuln.append(tuple(val.values()))
        for _, value in self.vulnmeta_data.items():
            self.lst_vuln_metadata.append(value['data'])
        # self.vuln_data = lst_vuln
        return lst_vuln, self.lst_vuln_metadata

    def get_nvd_data(self, allow_versions, cve_id, add_dict):
        """
        Method to get linux data
        :param allow_versions: allow_versions
        :param add_dict: add_dict
        :return: tuple
        """
        cpe_products = add_dict['total_cpe_products'][cve_id] if cve_id in add_dict['total_cpe_products'] else []
        packages = {}
        lst_vuln_data, lst_vuln_metadata = [], []
        for cpe_product in cpe_products:
            cpe_uri = cpe_product[0]
            cpe_detail = get_cpe_details(cpe_uri, is_application=True)
            cpe_uri_running_on = cpe_product[5]

            if cpe_detail:
                cpe_detail['cpe_uri'] = cpe_detail['cpe_uri'].replace('\\', '')
                cpe_detail['product'] = cpe_detail['product'].replace('\\', '').replace('<ANY>', '*').replace(
                    '<NA>', '-')
                cpe_detail['version'] = cpe_detail['version'].replace('\\', '').replace('<NA>', '')

                product_sw = \
                    f"{cpe_detail['vendor']}SEPARATOR{cpe_detail['product']}SEPARATOR{cpe_detail['target_sw']}" \
                    f"SEPARATOR{cpe_uri_running_on}"
                version = cpe_detail['version']
                version_with_ops = get_version_constraints(cpe_product, version)
                if product_sw not in packages:
                    packages[product_sw] = {'cpes': [], 'version_constraint': [], 'version_format': 'unknown',
                                            'proxy_vulnerabilities': '[]', 'related_vulnerabilities': 'null',
                                            'fixed_in_version': '', 'fixed_in_versions': 'null',
                                            'fix_state': 'unknown', 'advisories': 'null', 'package_qualifiers': []}

                if cpe_uri_running_on:
                    package_qualifier = {'kind': 'platform-cpe', 'cpe': cpe_uri_running_on}
                    if package_qualifier not in packages[product_sw]['package_qualifiers']:
                        packages[product_sw]['package_qualifiers'].append(package_qualifier)
                packages[product_sw]['cpes'].append(cpe_uri)
                packages[product_sw]['version_constraint'].extend(version_with_ops)

        table_col_dict = initialise_variables()
        table_col_dict['id'] = cve_id

        table_col_dict['record_source'] = 'nvdv2:cves'
        if allow_versions in ('v3', 'v4', 'v5'):
            table_col_dict['record_source'] = 'nvdv2:nvdv2:cves'

        table_col_dict['namespace'] = 'nvd'
        if allow_versions in ('v4', 'v5'):
            table_col_dict['namespace'] = 'nvd:cpe'
        table_col_dict['data_source'] = f'https://nvd.nist.gov/vuln/detail/{cve_id}'
        cvss, severity, cvss_v2_impact, cvss_v3_impact = get_nvd_cvss_data(add_dict['cve_detail'], allow_versions)
        cvss_v2_impact_ = None
        cvss_v3_impact_ = None
        if cvss_v2_impact:
            cvss_v2_impact_ = cvss_v2_impact['Metrics']
            cvss_v2_impact_['Vector'] = cvss_v2_impact['Vector']
        if cvss_v3_impact:
            cvss_v3_impact_ = cvss_v3_impact['Metrics']
            cvss_v3_impact_['Vector'] = cvss_v3_impact['Vector']
        table_col_dict['cvss_v2'] = cvss_v2_impact_
        table_col_dict['cvss_v3'] = cvss_v3_impact_
        table_col_dict['cvss'] = json.dumps(cvss) if cvss else '[]'
        table_col_dict['severity'] = severity
        table_col_dict['urls'] = json.dumps(add_dict['links']) if add_dict['links'] else 'null'
        table_col_dict['description'] = add_dict['cve_detail']['description'] if \
            add_dict['cve_detail']['description'] else ""
        if '** REJECT **' in str(table_col_dict['description']):
            return [], []
        table_col_dict['links'] = json.dumps(add_dict['links'])

        vuln_metadata_dict = {}
        if allow_versions in version_dict:
            meta_data_fields = version_dict[allow_versions]['vuln_meta_data']
            for fields in meta_data_fields:
                vuln_metadata_dict[fields] = table_col_dict[fields]
            vuln_meta_rec = table_col_dict['id'] + "::" + table_col_dict['namespace']
            if vuln_meta_rec not in self.vulnmeta_data:
                self.vulnmeta_data[vuln_meta_rec] = 1
            lst_vuln_metadata.append(tuple(vuln_metadata_dict.values()))

        for package_name, package_details in packages.items():

            package = re.sub(r"([+&()',/:!])", r'\\' + '\\1', package_name.split('SEPARATOR')[1], 0, re.MULTILINE)
            if cve_id in add_dict['cve_prod_map'] and package in add_dict['cve_prod_map'][cve_id] \
                    and add_dict['cve_prod_map'][cve_id][package]:
                created_at = add_dict['cve_prod_map'][cve_id][package].date()
            else:
                if cve_id in add_dict['cve_prod_on_os'] and package in add_dict['cve_prod_on_os'][cve_id] \
                        and add_dict['cve_prod_on_os'][cve_id][package]:
                    created_at = add_dict['cve_prod_on_os'][cve_id][package].date()
                else:
                    created_at = add_dict['cve_detail']['created_at'].date()

            tmp_package_qualifiers_for_uniq_key = package_name.split('SEPARATOR')[3]

            table_col_dict['package_name'] = package_name.split('SEPARATOR')[1]
            table_col_dict['version_constraint'] = ' || '.join(package_details['version_constraint']).strip()
            table_col_dict['version_format'] = package_details['version_format']
            table_col_dict['cpes'] = json.dumps(list(set(package_details['cpes'])), separators=(',', ':'))
            table_col_dict['proxy_vulnerabilities'] = package_details['proxy_vulnerabilities']
            table_col_dict['fixed_in_version'] = package_details['fixed_in_version']
            table_col_dict['related_vulnerabilities'] = package_details['related_vulnerabilities']
            table_col_dict['fixed_in_versions'] = package_details['fixed_in_versions']
            table_col_dict['fix_state'] = package_details['fix_state']
            table_col_dict['advisories'] = package_details['advisories']
            table_col_dict['created_at'] = str(created_at)
            table_col_dict['package_qualifiers'] = json.dumps(package_details['package_qualifiers']) \
                if package_details['package_qualifiers'] else None
            vuln_data_dict = {}
            if allow_versions in version_dict:
                vuln_data_fields = version_dict[allow_versions]['vuln_data']
                for fields in vuln_data_fields:
                    vuln_data_dict[fields] = table_col_dict[fields]
                vuln_rec = table_col_dict['id'] + "::" + table_col_dict['namespace'] + "::" + table_col_dict[
                    'package_name'] + "::" + table_col_dict['version_constraint'] + "::" + table_col_dict[
                    'cpes'] + "::" + tmp_package_qualifiers_for_uniq_key
                if vuln_rec not in self.vuln_data:
                    self.vuln_data[vuln_rec] = 1
                    lst_vuln_data.append(tuple(vuln_data_dict.values()))
        return lst_vuln_data, lst_vuln_metadata
